@@ -42,7 +42,7 @@
 // used to check if device is mounted
 #if defined(__LINUX__)
 #include <mntent.h>
-#elif defined (__BSD__)
+#elif defined (__BSD__) || defined (__OSX__)
 #include <sys/ucred.h>
 #include <sys/mount.h>
 #endif
@@ -85,7 +85,7 @@ int32_t check_mounted(char *filename) {
 
 	return ret;
 
-#elif defined(__BSD__)
+#elif defined(__BSD__) || defined(__OSX__)
 	struct statfs *mntbuf;
 	int i, mntsize;
 	int32_t ret = 0;
@@ -1096,7 +1096,7 @@ int32_t openFileSystem(char *path, uint32_t mode, struct sFileSystem *fs) {
 			switch (ret) {
 				case 0: break;  // filesystem not mounted
 				case 1:		// filesystem mounted
-					myerror("Filesystem is mounted!");
+					myerror("Filesystem is mounted. Please unmount!");
 					return -1;
 				case -1:	// unable to check
 				default:
@@ -1247,7 +1247,7 @@ int32_t openFileSystem(char *path, uint32_t mode, struct sFileSystem *fs) {
 
 #ifndef __WIN32__
 	// convert utf 16 le to local charset
-	fs->cd = iconv_open("//TRANSLIT", "UTF-16LE");
+	fs->cd = iconv_open("", "UTF-16LE");
 	if (fs->cd == (iconv_t)-1) {
 			myerror("iconv_open failed!");
 	return -1;
