@@ -61,8 +61,8 @@
 // maximum path len on FAT file systems (above specification)
 #define MAX_PATH_LEN 512
 
-// maximum filename length on exFAT
-#define MAX_EXFAT_FILENAME_LEN 255
+// maximum filename length
+#define MAX_FILENAME_LEN (uint32_t) 255
 
 // maximum file len
 // (specification: file < 4GB which is
