@@ -60,7 +60,7 @@ char *getCharSet(void) {
 	return NULL;
 }
 
-int32_t parseLongFilenamePart(struct sLongDirEntry *lde, char *str, iconv_t cd) {
+int32_t parseLongFilenamePart(struct sLongDirEntry *lde, char *utf16lestr) {
 /*
 	retrieves a part of a long filename from a
 	directory entry
@@ -68,64 +68,12 @@ int32_t parseLongFilenamePart(struct sLongDirEntry *lde, char *str, iconv_t cd)
 */
 
 	assert(lde != NULL);
-	assert(str != NULL);
+	assert(utf16lestr != NULL);
 
-	size_t incount;
-	size_t outcount;
-	char utf16str[28];
-
-	str[0]='\0';
-
-	memcpy(utf16str, (&lde->LDIR_Ord+1), 10);
-	memcpy(utf16str+10, (&lde->LDIR_Ord+14), 12);
-	memcpy(utf16str+22, (&lde->LDIR_Ord+28), 4);
-	memset(utf16str+26, 0, 2);
-
-	incount=26;
-	outcount=MAX_PATH_LEN;
-
-	int i;
-	for (i=0;i<12; i++) {
-		if ((utf16str[i*2] == '\0') && (utf16str[i*2+1] == '\0')) {
-			incount=i*2;
-			break;
-		}
-	}
-
-
-	//printf("Incount: %u\n", incount);
-
-#ifdef __WIN32__
-
-	if (WideCharToMultiByte(CP_ACP, 0, (LPCWSTR)utf16str, (int) incount, str, (int) outcount, NULL, NULL) == 0) {
-		stderror();
-		return -1;
-	}
-
-#else // Linux et al
-
-	char *outptr = &(str[0]);
-	char *inptr = &(utf16str[0]);
-	size_t ret;
-	
-	while (incount != 0) {
-		if ((ret=iconv(cd, &inptr, &incount, &outptr, &outcount)) == (size_t)-1) {
-			if (errno == EILSEQ) {
-				outptr[0]='?';
-				outptr++;
-				incount+=2;
-				outcount++;
-				inptr+=2;
-			} else {
-				stderror();
-				myerror("WARNING: iconv failed!");
-				break;
-			}
-		}
-	}
-  	outptr[0]='\0';
-
-#endif
+	memcpy(utf16lestr, (&lde->LDIR_Ord+1), 10);
+	memcpy(utf16lestr+10, (&lde->LDIR_Ord+14), 12);
+	memcpy(utf16lestr+22, (&lde->LDIR_Ord+28), 4);
+	memset(utf16lestr+26, 0, 2);
 
 	return 0;
 }
@@ -285,15 +233,15 @@ int32_t parseExFATClusterChain(struct sFileSystem *fs, struct sClusterChain *cha
 	uint32_t entries=0;
 	uint32_t expected_entries=0;
 	uint32_t r;
+	uint32_t nameLength=0;
 
 	struct sExFATDirEntry de;
 	struct sExFATDirEntrySet *des;
 	struct sExFATDirEntryList *del=NULL;
 
-	char name[MAX_PATH_LEN+1];
-	char str[31];
+	char utf16_filename[MAX_FILENAME_LEN*2+1];
+	char utf8_filename[MAX_FILENAME_LEN*4+1];
 	char *outptr, *inptr;
-	uint8_t nameLength=0;
 
 	size_t outcount=30;
 	size_t incount, iret;
@@ -304,7 +252,6 @@ int32_t parseExFATClusterChain(struct sFileSystem *fs, struct sClusterChain *cha
 
 	*reordered=0;
 
-	name[0]='\0';
 	while (chain != NULL) {
 		device_seekset(fs->device, getClusterOffset(fs, chain->cluster));
 		// fprintf(stderr, "cluster=%x;clusterOffset=%x\n", chain->cluster, getClusterOffset(fs, chain->cluster));
@@ -352,6 +299,12 @@ int32_t parseExFATClusterChain(struct sFileSystem *fs, struct sClusterChain *cha
 						return -1;
 					}
 					nameLength=de.entry.streamExtDirEntry.nameLen;
+					if (nameLength > MAX_FILENAME_LEN) {
+						myerror("Specified File name length in STEAM EXTENSION ENTRY is longer than %lu bytes (%lu bytes)!",
+								MAX_FILENAME_LEN, nameLength);
+						return -1;
+					}
+					
 					entries++;
 				} else if (!expected_entries) {
 					myerror("Secondary directory entries are not expected here (%u)!", ret);
@@ -370,46 +323,56 @@ int32_t parseExFATClusterChain(struct sFileSystem *fs, struct sClusterChain *cha
 					}
 					entries++;
 
-					// copy filename part to filename
-					outptr = &(str[0]);
-					str[0]='\0';
-					inptr=&(de.entry.FileNameExtDirEntry.filename[0]);
-					if (entries == expected_entries) { // last entry, so file name part is shorter
-						incount = 2 * (nameLength - (entries - 3) * 15);
-					} else {
-						incount=30;
+					if ((entries - 3) * 15 > nameLength) {
+						myerror("%u FILE NAME EXTENSION entries lead to filename length of more than %ul which is greater than specified in FILE STREAM EXTENSION entry!",
+								entries - 2, nameLength);
+						return -1;
 					}
 
-					//printf("incount: %d, length: %d\n", incount, nameLength);
-					outcount=MAX_PATH_LEN;
+					// copy filename part to filename
+					memcpy(utf16_filename+(entries - 3)*30,de.entry.FileNameExtDirEntry.filename, 30);
 
-					/*printf("file name part: " \
+					/*
+					printf("file name part: " \
 						"%02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx " \
 						"%02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx " \
 						"%02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx\n",
-						inptr[0], inptr[1], inptr[2], inptr[3], inptr[4],
-						inptr[5], inptr[6], inptr[7], inptr[8], inptr[9],
-						inptr[10], inptr[11], inptr[12], inptr[13], inptr[14],
-						inptr[15], inptr[16], inptr[17], inptr[18], inptr[19],
-						inptr[20], inptr[21], inptr[22], inptr[23], inptr[24],
-						inptr[25], inptr[26], inptr[27], inptr[28], inptr[29]
-					);*/
-
-					while (incount != 0) {
-				                if ((iret=iconv(fs->cd, &inptr, &incount, &outptr, &outcount)) == (size_t)-1) {
-							stderror();
-				                        myerror("iconv failed! %d", iret);
-							return -1;
-				                }
-				        }
-					outptr[0]='\0';
-
-					strncat(name, str, 31);
+						de.entry.FileNameExtDirEntry.filename[0], de.entry.FileNameExtDirEntry.filename[1],
+						de.entry.FileNameExtDirEntry.filename[2], de.entry.FileNameExtDirEntry.filename[3],
+						de.entry.FileNameExtDirEntry.filename[4], de.entry.FileNameExtDirEntry.filename[5],
+						de.entry.FileNameExtDirEntry.filename[6], de.entry.FileNameExtDirEntry.filename[7],
+						de.entry.FileNameExtDirEntry.filename[8], de.entry.FileNameExtDirEntry.filename[9],
+						de.entry.FileNameExtDirEntry.filename[10], de.entry.FileNameExtDirEntry.filename[11],
+						de.entry.FileNameExtDirEntry.filename[12], de.entry.FileNameExtDirEntry.filename[13],
+						de.entry.FileNameExtDirEntry.filename[14], de.entry.FileNameExtDirEntry.filename[15],
+						de.entry.FileNameExtDirEntry.filename[16], de.entry.FileNameExtDirEntry.filename[17],
+						de.entry.FileNameExtDirEntry.filename[18], de.entry.FileNameExtDirEntry.filename[19],
+						de.entry.FileNameExtDirEntry.filename[20], de.entry.FileNameExtDirEntry.filename[21],
+						de.entry.FileNameExtDirEntry.filename[22], de.entry.FileNameExtDirEntry.filename[23],
+						de.entry.FileNameExtDirEntry.filename[24], de.entry.FileNameExtDirEntry.filename[25],
+						de.entry.FileNameExtDirEntry.filename[26], de.entry.FileNameExtDirEntry.filename[27],
+						de.entry.FileNameExtDirEntry.filename[28], de.entry.FileNameExtDirEntry.filename[29]
+					);
+					*/
 
 					// we are done here
 					if (entries == expected_entries) {
 
-						des=newExFATDirEntrySet(name, del, entries);
+						// convert utf-16 string form FILE NAME EXTENSION entries to utf-18 string in current locale
+						incount = nameLength*2;
+						outcount = MAX_FILENAME_LEN*4;
+						inptr = &(utf16_filename[0]);
+						outptr = &(utf8_filename[0]);
+						while (incount != 0) {
+						        if ((iret=iconv(fs->cd, &inptr, &incount, &outptr, &outcount)) == (size_t)-1) {
+								stderror();
+						                myerror("iconv failed! %d", iret);
+								return -1;
+						        }
+						}
+						outptr[0]='\0';				
+												
+						des=newExFATDirEntrySet(utf8_filename, del, entries);
 						if (!des) {
 							myerror("Could not create exFAT directory entry set");
 							return -1;
@@ -433,7 +396,7 @@ int32_t parseExFATClusterChain(struct sFileSystem *fs, struct sClusterChain *cha
 
 						(*direntrysets)++;
 						entries=0;
-						name[0]='\0';
+
 					}
 				} else if (entries >= expected_entries) {
 					myerror("Too many file name extension directory entries!");
@@ -540,7 +503,8 @@ int32_t parseClusterChain(struct sFileSystem *fs, struct sClusterChain *chain, s
 	union sDirEntry de;
 	struct sDirEntryList *lnde;
 	struct sLongDirEntryList *llist;
-	char tmp[MAX_PATH_LEN+1], dummy[MAX_PATH_LEN+1], sname[MAX_PATH_LEN+1], lname[MAX_PATH_LEN+1];
+	char tmp[28], utf16le_lname[MAX_PATH_LEN*2+2], sname[MAX_PATH_LEN+1], lname[MAX_PATH_LEN*4+1], dummy[MAX_PATH_LEN*2+1];
+	size_t incount, outcount;
 
 	*direntries=0;
 
@@ -549,12 +513,21 @@ int32_t parseClusterChain(struct sFileSystem *fs, struct sClusterChain *chain, s
 	llist = NULL;
 	lname[0]='\0';
 	*reordered=0;
+	
+	utf16le_lname[MAX_PATH_LEN*4]='\0';
+	utf16le_lname[MAX_PATH_LEN*4+1]='\0';
+	
 	while (chain != NULL) {
 		device_seekset(fs->device, getClusterOffset(fs, chain->cluster));
 		for (j=0;j<fs->maxDirEntriesPerCluster;j++) {
 			entries++;
 			ret=parseEntry(fs, &de);
-
+			
+			/*
+			printf("entry: %u\n", ret);
+			printf("entries: %u\n", entries);
+			*/
+			
 			switch(ret) {
 			case -1:
 				myerror("Failed to parse directory entry!");
@@ -586,6 +559,47 @@ int32_t parseClusterChain(struct sFileSystem *fs, struct sClusterChain *chain, s
 					printf("!%s (#%s)\n", (lname[0] != '\0') ? lname : "n/a", sname+1);
 				}
 */
+
+				if (entries>1) { // we have a long filename
+					// count characters in long filename
+					uint32_t filenameLen=0;
+					char *p=utf16le_lname;
+					while((p[0] != '\0') || (p[1] != '\0')) {
+						p+=2;filenameLen++;
+					}
+					
+					if (filenameLen > MAX_FILENAME_LEN) {
+						myerror("Filename length exceeds %lu characters (%lu characters)!", MAX_FILENAME_LEN, filenameLen);
+						return -1;
+					}
+				}
+
+				// convert utf-16le long filename to utf-8 in current locale
+				lname[0]='\0';
+				char *outptr = &(lname[0]);
+				char *inptr = &(utf16le_lname[0]);
+				incount=(entries-1)*26;
+				outcount=MAX_PATH_LEN*4;
+
+				while (incount != 0) {
+					if ((iconv(fs->cd, &inptr, &incount, &outptr, &outcount)) == (size_t)-1) {
+						if (errno == EILSEQ) {
+							outptr[0]='?';
+							outptr++;
+							incount-=2;
+							outcount++;
+							inptr+=2;
+						} else {
+							stderror();
+							myerror("iconv failed (errno %u)!", errno);
+							return -1;
+						}
+					}
+				}
+			  	outptr[0]='\0';
+			  	
+			  	//printf("lname (sname): %s (%s)\n", lname, sname);
+
 				lnde=newDirEntry(sname, lname, &de.ShortDirEntry, llist, entries);
 				if (lnde == NULL) {
 					myerror("Failed to create DirEntry!");
@@ -602,10 +616,12 @@ int32_t parseClusterChain(struct sFileSystem *fs, struct sClusterChain *chain, s
 				(*direntries)++;
 				entries=0;
 				llist = NULL;
-				lname[0]='\0';
+				utf16le_lname[0]=L'\0';
+				sname[0]='\0';
 				break;
 			case 2: // long dir entry
-				if (parseLongFilenamePart(&de.LongDirEntry, tmp, fs->cd)) {
+			
+				if (parseLongFilenamePart(&de.LongDirEntry, tmp)) {
 					myerror("Failed to parse long filename part!");
 					return -1;
 				}
@@ -616,12 +632,17 @@ int32_t parseClusterChain(struct sFileSystem *fs, struct sClusterChain *chain, s
 					myerror("Failed to insert LongDirEntry!");
 					return -1;
 				}
-
-				strncpy(dummy, tmp, MAX_PATH_LEN);
-				dummy[MAX_PATH_LEN]='\0';
-				strncat(dummy, lname, MAX_PATH_LEN - strlen(dummy));
-				dummy[MAX_PATH_LEN]='\0';
-				strncpy(lname, dummy, MAX_PATH_LEN+1);
+				
+				/*printf("tmp: ");
+				for(int i=0;i<26;i++){
+					printf("%02hhx ", tmp[i]);
+				}
+				printf("\n");*/
+				
+				memcpy(dummy, tmp, 26);
+				memcpy(dummy+26, utf16le_lname, (entries -1)*26);
+				memcpy(utf16le_lname, dummy, entries*26);
+				memset(utf16le_lname+entries*26, 0, 2);
 
 				break;
 			default:
@@ -657,7 +678,8 @@ int32_t parseFat1xRootDirEntries(struct sFileSystem *fs, struct sDirEntryList *l
 	union sDirEntry de;
 	struct sDirEntryList *lnde;
 	struct sLongDirEntryList *llist;
-	char tmp[MAX_PATH_LEN+1], dummy[MAX_PATH_LEN+1], sname[MAX_PATH_LEN+1], lname[MAX_PATH_LEN+1];
+	char tmp[28], utf16le_lname[MAX_PATH_LEN*2+1], sname[MAX_PATH_LEN+1], lname[MAX_PATH_LEN*4+1], dummy[MAX_PATH_LEN*2+1];
+	size_t incount, outcount;
 
 	*direntries=0;
 
@@ -704,6 +726,47 @@ int32_t parseFat1xRootDirEntries(struct sFileSystem *fs, struct sDirEntryList *l
 				printf("!%s (#%s)\n", (lname[0] != '\0') ? lname : "n/a", sname+1);
 			}
 */
+
+			if (entries>1) { // we have a long filename
+				// count characters in long filename
+				uint32_t filenameLen=0;
+				char *p=utf16le_lname;
+				while((p[0] != '\0') || (p[1] != '\0')) {
+					p+=2;filenameLen++;
+				}
+				
+				if (filenameLen > MAX_FILENAME_LEN) {
+					myerror("Filename length exceeds %lu characters (%lu characters)!", MAX_FILENAME_LEN, filenameLen);
+					return -1;
+				}
+			}
+
+			// convert utf-16le long filename to utf-8 in current locale
+			lname[0]='\0';
+			char *outptr = &(lname[0]);
+			char *inptr = &(utf16le_lname[0]);
+			incount=(entries-1)*26;
+			outcount=MAX_PATH_LEN*4;
+
+			while (incount != 0) {
+				if ((iconv(fs->cd, &inptr, &incount, &outptr, &outcount)) == (size_t)-1) {
+					if (errno == EILSEQ) {
+						outptr[0]='?';
+						outptr++;
+						incount-=2;
+						outcount++;
+						inptr+=2;
+					} else {
+						stderror();
+						myerror("iconv failed (errno %u)!", errno);
+						return -1;
+					}
+				}
+			}
+		  	outptr[0]='\0';
+		  	
+		  	//printf("lname (sname): %s (%s)\n", lname, sname);
+
 			lnde=newDirEntry(sname, lname, &de.ShortDirEntry, llist, entries);
 			if (lnde == NULL) {
 				myerror("Failed to create DirEntry!");
@@ -720,10 +783,11 @@ int32_t parseFat1xRootDirEntries(struct sFileSystem *fs, struct sDirEntryList *l
 			(*direntries)++;
 			entries=0;
 			llist = NULL;
-			lname[0]='\0';
+			utf16le_lname[0]=L'\0';
+			sname[0]='\0';
 			break;
 		case 2: // long dir entry
-			if (parseLongFilenamePart(&de.LongDirEntry, tmp, fs->cd)) {
+			if (parseLongFilenamePart(&de.LongDirEntry, tmp)) {
 				myerror("Failed to parse long filename part!");
 				return -1;
 			}
@@ -735,12 +799,16 @@ int32_t parseFat1xRootDirEntries(struct sFileSystem *fs, struct sDirEntryList *l
 				return -1;
 			}
 
-			strncpy(dummy, tmp, MAX_PATH_LEN);
-			dummy[MAX_PATH_LEN]='\0';
-			strncat(dummy, lname, MAX_PATH_LEN - strlen(dummy));
-			dummy[MAX_PATH_LEN]='\0';
-			strncpy(lname, dummy, MAX_PATH_LEN+1);
-			//dummy[MAX_PATH_LEN]='\0';
+			/*printf("tmp: ");
+			for(int i=0;i<26;i++){
+				printf("%02hhx ", tmp[i]);
+			}
+			printf("\n");*/
+			
+			memcpy(dummy, tmp, 26);
+			memcpy(dummy+26, utf16le_lname, (entries -1)*26);
+			memcpy(utf16le_lname, dummy, entries*26);
+			memset(utf16le_lname+entries*26, 0, 2);
 			break;
 		default:
 			myerror("Unhandled return code!");
