@@ -78,8 +78,7 @@ int32_t addStringToStringList(struct sStringList *stringList, const char *str) {
 		return -1;
 	}
 
-	strncpy(stringList->next->str, str, len);
-	stringList->next->str[len] = '\0';
+	memcpy(stringList->next->str, str, len+1);
 
 	return 0;
 
